"""
An extended commonmark renderer based on markdown-it-py.
"""
from __future__ import annotations

import sys
import typing as t

from markdown_it.common.utils import escapeHtml, unescapeAll
from markdown_it.renderer import RendererHTML

if sys.version_info < (3, 11):
    import tomli as tomllib
else:
    import tomllib

if t.TYPE_CHECKING:
    from collections.abc import Sequence
    from markdown_it.token import Token
    from markdown_it.utils import EnvType, OptionsDict


def get_container_renderer(container_name: str, html_tag: str):
    """
    Factory function for markdown container renderers.

    Unnecessary internal `<p>` wrapper tags will be eliminated.

    :param container_name: The name of the container. Will be included as a
        `class` in the output HTML container.
    :param html_tag: The HTML tag for the container.
    """
    def render(
        self: RendererHTML,
        tokens: Sequence[Token],
        idx: int,
        _options: OptionsDict,
        env: EnvType,
    ) -> str:
        tokens[idx].tag = html_tag
        # add a class to the opening tag
        if tokens[idx].nesting == 1:
            tokens[idx].attrJoin("class", container_name)
            next_token = tokens[idx+1]
            if next_token.type == 'paragraph_open':
                next_token.hidden = True
                counter = idx + 2
                while tokens[counter].type != 'paragraph_close' or tokens[counter].level != next_token.level:
                    counter += 1
                tokens[counter].hidden = True

        return self.renderToken(tokens, idx, _options, env)

    render.__name__ = f'render_{container_name}_to_{html_tag}'
    return render


class AnchovyRendererHTML(RendererHTML):
    """
    A customized markdown-it-py HTML renderer, with hooks for better pygments
    integration and toml frontmatter support.
    """
    # https://github.com/executablebooks/markdown-it-py/issues/256
    def fence(self, tokens: Sequence[Token], idx: int, options: OptionsDict, _env: EnvType):
        """
        Handles rendering a markdown code fence, with optional syntax
        highlighting.
        """
        token = tokens[idx]
        info = unescapeAll(token.info).strip() if token.info else ''
        lang_name = info.split(maxsplit=1)[0] if info else ''

        return (
            options.highlight
            and options.highlight(token.content, lang_name, '')
            or escapeHtml(token.content)
        )

    def front_matter(self, tokens: Sequence[Token], idx: int, _options: OptionsDict, env: EnvType):
        """
        Handles parsing markdown frontmatter using TOML.
        """
        parsed = tomllib.loads(tokens[idx].content)
        env['anchovy_meta'].update(parsed)
        return ''



"""
Simple Steps and a base class for Steps that invoke external commandline tools.
"""
from __future__ import annotations

import abc
import shutil
import subprocess
import typing as t
from pathlib import Path

from .core import Step

if t.TYPE_CHECKING:
    from _typeshed import StrOrBytesPath


class DirectCopyStep(Step):
    """
    A simple Step which only copies a file to the output directory without
    renaming or extension changes.
    """
    def __call__(self, path: Path, output_paths: list[Path]):
        for target_path in output_paths:
            target_path.parent.mkdir(parents=True, exist_ok=True)
            shutil.copy(path, target_path)


class BaseCommandStep(Step):
    """
    A base class for steps that run an external command to generate a file.
    """
    @abc.abstractmethod
    def get_command(self, input_path: Path, output_path: Path) -> StrOrBytesPath | list[StrOrBytesPath]:
        """
        Abstract method that must return a commandline ready for subprocess.
        """

    def group_outputs(self, output_paths: list[Path]) -> list[list[Path]]:
        """
        Overridable method which determines which outputs must be generated
        separately. Default behavior groups outputs by extension.
        """
        groups = {}
        for path in output_paths:
            groups.setdefault(path.suffix, []).append(path)
        return list(groups.values())

    def __call__(self, path: Path, output_paths: list[Path]):
        for egroup in self.group_outputs(output_paths):
            first = None
            for opath in egroup:
                opath.parent.mkdir(parents=True, exist_ok=True)
                if first:
                    shutil.copy(first, opath)
                else:
                    subprocess.check_output(
                        self.get_command(path, opath),
                        stderr=subprocess.STDOUT
                    )
                    first = opath



"""
Explicit chain of custody management and intelligent rebuilds for Anchovy.
"""
from __future__ import annotations

import hashlib
import json
import typing as t
from importlib.metadata import version
from pathlib import Path

from .pretty_utils import print_with_style

if t.TYPE_CHECKING:
    from collections.abc import Sequence
    from .core import Context, ContextDir


_JsonSerializable: t.TypeAlias = 'str | int | float | bool | None | _JsonDict | Sequence[_JsonSerializable]'
_JsonDict = dict[str, _JsonSerializable]


def checksum(path: Path, hashname: str = 'sha1', _bufsize=2**18):
    """
    Calculate a checksum for a `Path`. Directories result in empty checksums.
    """
    if path.is_dir():
        return ''
    digest = hashlib.new(hashname)

    buf = bytearray(_bufsize)
    view = memoryview(buf)
    with path.open('rb') as file:
        while True:
            size = file.readinto(buf)
            if size == 0:
                break  # EOF
            digest.update(view[:size])

    return digest.hexdigest()


class CustodyEntry:
    """
    Class holding custody info for a single input or output path.
    """
    def __init__(self, entry_type: str, key: str, meta: dict | None = None):
        self.entry_type = entry_type
        self.key = key
        self.meta = meta or {}

    def __str__(self):
        return f'{self.entry_type}:{self.key}'

    def __getitem__(self, key):
        return self.meta[key]


class Custodian:
    """
    Class for managing custody info and intelligent rebuilds.
    """
    encoding = 'utf-8'
    context: 'Context'

    def __init__(self,
                 parameters: _JsonDict | None = None,
                 info: dict[str, str] | None = None):
        self.checkers: dict[str, t.Callable[[CustodyEntry], bool]] = {'path': self.check_path}

        self.parameters: _JsonDict = {'anchovy_version': version('anchovy')}
        if parameters:
            self.parameters.update(parameters)
        self.prior_parameters: _JsonDict = {}
        self.stale_parameters = True

        self.info = info or {}

        # output_key: {input_key: [sibling_keys]}
        self.graph: dict[str, dict[str, list[str]]] = {}
        self.prior_graph: dict[str, dict[str, list[str]]] = {}

        # key: (type, meta) (i.e. the parts of a CustodyEntry)
        self.meta: dict[str, tuple[str, _JsonDict]] = {}
        self.prior_meta: dict[str, tuple[str, _JsonDict]] = {}

    def bind(self, context: 'Context'):
        """
        Bind this `Custodian` to a `Context` and update info using the
        `Context`'s settings.
        """
        self.context = context
        for key in context.settings:
            if key not in self.info:
                self.info[key] = str(context.settings[key])

    # Default behaviors for Paths

    def genericize_path(self, path: Path):
        """
        Genericize a path, removing run-specific folder references and
        converting to a key.
        """
        for dir_key in ('input_dir', 'output_dir', 'working_dir'):
            parent = self.context[dir_key]
            if path.is_relative_to(parent):
                path = dir_key / path.relative_to(parent)
                break
        return path.as_posix()

    def degenericize_path(self, key: str):
        """
        Undo `genericize_path()` to turn a key back into a Path.
        """
        path = Path(key)
        # -2 because path.parents walks upward (parents[0] is the same as
        # path.parent) and -1 is the root (in our case, '.')
        dir_key = t.cast('ContextDir', str(path.parents[-2]))
        return self.context[dir_key] / path.relative_to(dir_key)

    def get_all_paths(self):
        """
        Generator of every output key in the graph as a Path.
        """
        return (self.degenericize_path(key) for key in self.graph)

    def entry_from_path(self, path: Path):
        """
        Create a `CustodyEntry` for a path. Provides a sha1 checksum along with
        stat modified time and size for path checkers.
        """
        # The default checker only looks at a checksum, but it's not much added
        # cost to also stat, and doing so means that an end user can switch to
        # m_time testing by registering a new checker and does not need to
        # subclass CustodyManager for this common case.
        stat = path.stat()
        meta = {'sha1': checksum(path), 'm_time': stat.st_mtime, 'size': stat.st_size}
        return CustodyEntry('path', self.genericize_path(path), meta)

    def check_path(self, entry: CustodyEntry) -> bool:
        """
        Default sha1-based checker for path staleness.
        """
        return entry['sha1'] == checksum(self.degenericize_path(entry.key))

    def ensure_entry(self, record: Path | CustodyEntry):
        """
        Create a `CustodyEntry` if the parameter is a Path, otherwise just
        return the existing `CustodyEntry`.
        """
        if isinstance(record, CustodyEntry):
            return record
        return self.entry_from_path(record)

    def register_checker(self, entry_type: str, override: bool = True):
        """
        Decorator for registering a staleness checker for a specific type of
        `CustodyEntry`.
        """
        def register(func: t.Callable[[CustodyEntry], bool]):
            if override or entry_type not in self.checkers:
                self.checkers[entry_type] = func
        return register

    def load_file(self, path: Path):
        """
        Load prior custody data and metadata from a JSON file and evaluate
        parameter staleness.
        """
        if not path.exists():
            return
        data = json.loads(path.read_text(self.encoding))
        self.prior_parameters = data['parameters']
        self.stale_parameters = self.parameters != self.prior_parameters
        self.prior_graph = data['graph']
        self.prior_meta = data['meta']

    def dump_file(self, path: Path):
        """
        Dump all custody data and metadata from the current run into a JSON
        file.
        """
        data = {
            'info': self.info,
            'parameters': self.parameters,
            'graph': self.graph,
            'meta': self.meta,
        }
        with path.open('w', encoding=self.encoding) as file:
            json.dump(data, file, indent=2)

    def update_meta(self, entry: CustodyEntry):
        """
        Store a `CustodyEntry` in serializable form.
        """
        self.meta[entry.key] = (entry.entry_type, entry.meta)

    def add_step(self,
                 sources: Sequence[Path | CustodyEntry],
                 outputs: Sequence[Path],
                 stale_msg: str):
        """
        Mark a Step as run, updating custody data and logging accordingly.
        """
        self.log_step(sources, outputs, stale=True, stale_msg=stale_msg)

        keys = []
        for o_entry in map(self.ensure_entry, outputs):
            keys.append(o_entry.key)
            self.update_meta(o_entry)

        for i_entry in map(self.ensure_entry, sources):
            self.update_meta(i_entry)
            for o_key in keys:
                self.graph.setdefault(o_key, {})[i_entry.key] = keys

    def skip_step(self, source: Path, outputs: list[Path]):
        """
        Mark a Step as skipped, updating custody data and logging accordingly.
        """
        i_entry = self.entry_from_path(source)
        g_output = self.genericize_path(outputs[0])
        prior_outputs = [self.degenericize_path(p) for p in self.prior_graph[g_output][i_entry.key]]

        self.log_step([source], prior_outputs, stale=False)

        self.update_meta(i_entry)
        for o_entry in map(self.entry_from_path, prior_outputs):
            self.update_meta(o_entry)
            self.graph.setdefault(o_entry.key, {}).update(self.prior_graph[o_entry.key])
            for s_key in self.prior_graph[o_entry.key]:
                self.meta.setdefault(s_key, self.prior_meta[s_key])
        return prior_outputs

    def log_step(self,
                 sources: Sequence[Path | CustodyEntry],
                 outputs: Sequence[Path],
                 *,
                 stale: bool = True,
                 stale_msg: str = ''):
        """
        Log custody information for a step according to its staleness.
        """
        if len(sources) == 1:
            msg = f'{sources[0]} ⇒ {", ".join(str(p) for p in outputs)}'
        else:
            msg = ''.join([
                '{\n\t',
                ',\n\t'.join(str(s) for s in sources),
                '\n} ⇒ {\n\t',
                ',\n\t'.join(str(p) for p in outputs),
                '\n}',
            ])
        if stale:
            print_with_style(f'{stale_msg}...\n{msg}')
        else:
            print_with_style('Skipped', msg, style='yellow')

    def check_prior(self, key: str):
        """
        Check whether the current resource corresponding to the given key
        matches its historical fingerprint.

        :param key: A `CustodyEntry.key`.
        """
        try:
            ptype, pmeta = self.prior_meta[key]
        except KeyError:
            return False

        try:
            checker = self.checkers[ptype]
        except KeyError as e:
            raise KeyError(f'No checker found for type {ptype!r}!') from e

        return checker(CustodyEntry(ptype, key, pmeta))

    def find_upstream(self, paths: list[Path]):
        """
        Find all input paths one step upstream of any of the specified paths.

        Relies only on historical data; combine with the
        `Custodian.stale_parameters` attribute and the `Custodian.check_prior()`
        method to ensure the historical data is still accurate.
        """
        return {
            s
            for p in paths
            for s in self.prior_graph.get(self.genericize_path(p), ())
        }

    def refresh_needed(self, source: Path, outputs: list[Path]):
        """
        Determines whether a refresh is needed for the result of processing a
        pair of Step parameters.

        :return: Whether the step should be rerun and a message explaining why
            or why not.
        """
        if self.stale_parameters:
            return True, 'Stale parameters'
        for path in outputs:
            if not path.exists():
                return True, f'Missing output ({path})'
        upstreams = self.find_upstream(outputs)
        if self.genericize_path(source) not in upstreams:
            return True, f'Missing upstream record ({source})'
        for up_key in upstreams:
            if not self.check_prior(up_key):
                return True, f'Stale upstream ({up_key})'
        return False, 'Up to date'



"""
Steps for reducing the load cost of webpages by combining and minifying
resources.
"""
import shutil
from pathlib import Path
from collections.abc import Sequence

from .core import ContextDir, Step
from .dependencies import PipDependency


class ResourcePackerStep(Step):
    """
    A simple resource packing Step, using a config file with a list of files to
    join into one.
    """
    encoding = 'utf-8'

    def __init__(self, source_dir: ContextDir = 'input_dir'):
        self.source_dir: ContextDir = source_dir

    def __call__(self, path: Path, output_paths: list[Path]):
        parent_dir = self.context[self.source_dir]
        input_paths = [
            (parent_dir / cleaned)
            for filename in path.open()
            if (cleaned := filename.strip()) and not cleaned.startswith('#')
        ]
        data = '\n\n'.join(f.read_text(self.encoding) for f in input_paths)

        for o_path in output_paths:
            o_path.parent.mkdir(parents=True, exist_ok=True)
        output_paths[0].write_text(data, self.encoding)
        for o_path in output_paths[1:]:
            shutil.copy(output_paths[0], o_path)

        input_paths.insert(0, path)
        return input_paths, output_paths


class CSSMinifierStep(Step):
    """
    A powerful CSS minification Step, using lightningcss to offer intelligent
    CSS reduction based on browsers supported and unused styles.
    """
    encoding = 'utf-8'

    @classmethod
    def get_dependencies(cls):
        return {
            PipDependency('lightningcss')
        }

    def __init__(self,
                 error_recovery: bool = False,
                 parser_flags: dict[str, bool] | None = None,
                 unused_symbols: set[str] | None = None,
                 browsers_list: Sequence[str] | None = ('defaults',),
                 minify: bool = True):

        self.error_recovery = error_recovery
        self.parser_flags = parser_flags or {}
        self.unused_symbols = unused_symbols
        self.browsers_list = list(browsers_list) if browsers_list else None
        self.minify = minify

    def __call__(self, path: Path, output_paths: list[Path]):
        import lightningcss
        data = lightningcss.process_stylesheet(
            path.read_text(self.encoding),
            filename=str(path),
            error_recovery=self.error_recovery,
            parser_flags=lightningcss.calc_parser_flags(**self.parser_flags),
            unused_symbols=self.unused_symbols,
            browsers_list=self.browsers_list,
            minify=self.minify
        )
        for o_path in output_paths:
            o_path.parent.mkdir(parents=True, exist_ok=True)
        output_paths[0].write_text(data, self.encoding)
        for o_path in output_paths[1:]:
            shutil.copy(output_paths[0], o_path)


class HTMLMinifierStep(Step):
    """
    A simple but fast HTML minification Step.
    """
    encoding = 'utf-8'
    minify_css = False
    minify_js = False

    @classmethod
    def get_dependencies(cls):
        return {
            (
                PipDependency('minify-html-onepass', check_name='minify_html_onepass')
                | PipDependency('minify-html', check_name='minify_html')
            ),
        }

    def __call__(self, path: Path, output_paths: list[Path]):
        params = {'minify_css': self.minify_css, 'minify_js': self.minify_js}
        try:
            from minify_html_onepass import minify
        except ImportError:
            from minify_html import minify
            params |= {
                'do_not_minify_doctype': True,
                'ensure_spec_compliant_unquoted_attribute_values': True,
                'keep_spaces_between_attributes': True,
            }
        data = minify(path.read_text(self.encoding), **params)
        for o_path in output_paths:
            o_path.parent.mkdir(parents=True, exist_ok=True)
        output_paths[0].write_text(data, self.encoding)
        for o_path in output_paths[1:]:
            shutil.copy(output_paths[0], o_path)



from pathlib import Path

from anchovy import (
    DirectCopyStep,
    InputBuildSettings,
    JinjaMarkdownStep,
    OutputDirPathCalc,
    REMatcher,
    Rule,
)


# Optional, and can be overridden with CLI arguments.
SETTINGS = InputBuildSettings(
    input_dir=Path(__file__).parent / 'basic_site',
    working_dir=Path('working/basic_site'),
    output_dir=Path('output/basic_site'),
    custody_cache=Path('output/basic_site.json'),
)
RULES = [
    # Ignore dotfiles found in either the input_dir or the working dir.
    Rule(
        (
            REMatcher(r'(.*/)*\..*', parent_dir='input_dir')
            | REMatcher(r'(.*/)*\..*', parent_dir='working_dir')
        ),
        None
    ),
    # Render markdown files, then stop processing them.
    Rule(
        REMatcher(r'.*\.md'),
        [OutputDirPathCalc('.html'), None],
        JinjaMarkdownStep()
    ),
    # Copy everything else in static/ directories through.
    Rule(
        REMatcher(r'(.*/)*static/.*', parent_dir='input_dir'),
        OutputDirPathCalc(),
        DirectCopyStep()
    ),
]



"""
This is the toolkit for Anchovy's own CLI, but offers an accessible API for
building project-specific CLIs.
"""
from __future__ import annotations

import argparse
import contextlib
import importlib
import sys
import tempfile
import typing as t
from pathlib import Path

from .core import BuildSettings, Context, InputBuildSettings, Rule, Step, StepUnavailableException
from .custody import Custodian
from .pretty_utils import print_with_style


class BuildNamespace:
    """
    Internal used to preserve typing between InputBuildSettings, argparse, and
    BuildSettings.
    """
    input_dir: Path
    output_dir: Path
    working_dir: Path | None
    custody_cache: Path | None
    purge_dirs: bool | None

    def __init__(self, settings: InputBuildSettings | None = None):
        if settings:
            self.__dict__.update(settings)

    def to_build_settings(self, resolved_working_dir: Path):
        """
        Convert this argparse-oriented namespace into a Context-ready
        BuildSettings.
        """
        purge_dirs = self.purge_dirs
        if purge_dirs is None and self.custody_cache is None:
            purge_dirs = True
        return BuildSettings(
            input_dir=self.input_dir,
            output_dir=self.output_dir,
            working_dir=resolved_working_dir,
            custody_cache=self.custody_cache,
            purge_dirs=purge_dirs
        )


@contextlib.contextmanager
def _wrap_temp(path: Path | None):
    if path:
        yield path
    else:
        with tempfile.TemporaryDirectory() as temp_dir:
            yield Path(temp_dir)


def parse_settings_args(settings: InputBuildSettings | None = None, argv: list[str] | None = None, **kw):
    """
    Internal function used by `run_from_rules()` to combine an instance of
    InputBuildSettings with CLI arguments to produce a BuildNamespace, which
    can be easily turned into BuildSettings.
    """
    namespace = BuildNamespace(settings)

    parser = argparse.ArgumentParser(**kw)
    parser.add_argument('-i', '--input',
                        help='input directory with raw files to process',
                        type=Path,
                        dest='input_dir',
                        default=Path('site'))
    parser.add_argument('-o', '--output',
                        help='output directory for final built files',
                        type=Path,
                        dest='output_dir',
                        default=Path('output'))

    group = parser.add_mutually_exclusive_group()
    group.add_argument('-w', '--working',
                       help='directory for intermediate files; defaults to a new temporary directory',
                       type=Path,
                       dest='working_dir')
    group.add_argument('--use-temporary',
                       help='force use of a temporary directory for intermediate files',
                       action='store_const',
                       dest='working_dir',
                       const=None)

    parser.add_argument('--custody-cache',
                        help='path to a cache file for chain of custody and change detection',
                        type=Path,
                        default=None)
    parser.add_argument('--purge',
                        help='purge the output directory before building',
                        action=argparse.BooleanOptionalAction,
                        dest='purge_dirs',
                        default=None)

    return parser.parse_args(argv, namespace=namespace)


def run_from_rules(settings: InputBuildSettings | None,
                   rules: list[Rule],
                   custodian: Custodian | None = None,
                   context_cls: t.Type[Context] = Context,
                   **kw):
    """
    Build a new Context from Settings, Rules, and command line arguments. Then,
    execute a build using the new Context. A Custodian and a custom Context
    class may be additionally supplied.
    """
    final_settings = parse_settings_args(settings, **kw)
    with _wrap_temp(final_settings.working_dir) as working_dir:
        context = context_cls(final_settings.to_build_settings(working_dir), rules, custodian)
        context.run()


def pprint_step(step: t.Type[Step]):
    """
    Prettily display dependency information for the given Step class.
    """
    missing = [
        str(d) for d in step.get_dependencies()
        if d.needed and not d.satisfied

    ]
    if missing:
        text = ', '.join(missing)
        print_with_style(f'✗ {step.__name__} (missing: {text})', style='red')
    else:
        print_with_style(f'✓ {step.__name__}', style='green')


def pprint_missing_deps(step: Step):
    """
    Prettily display an error for the given Step with missing dependencies.
    """
    print_with_style(
        f'{step} is unavailable due to missing dependencies!',
        file='stderr',
        style='red'
    )
    for dep in step.get_dependencies():
        missing = False
        if not dep.needed:
            style = None
        elif dep.satisfied:
            style = 'green'
        else:
            missing = True
            style = 'red'

        text = f'✗ {dep}: {dep.install_hint}' if missing else f'✓ {dep}'
        print_with_style(text, style=style)


def main():
    """
    Anchovy main function. Finds or creates a Context using an Anchovy config
    file and command line arguments, then executes a build using it.
    """
    parser = argparse.ArgumentParser(description='Build an anchovy project.')
    parser.add_argument('--audit-steps',
                        help=('show information about available, unavailable, '
                              'and used steps, instead of building the project'),
                        action='store_true')
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('-m',
                       help='import path of a config file to build',
                       type=importlib.import_module,
                       dest='module',
                       default=None)
    group.add_argument('config_file',
                       nargs='?',
                       help='file path to a config file to build',
                       type=Path,
                       default=None)

    args, remaining = parser.parse_known_args()

    if args.config_file:
        label: str = str(args.config_file)

        namespace: dict[str, t.Any] = {'__file__': label}
        with open(args.config_file, encoding='utf-8') as file:
            # We're basically emulating the python command line here.
            # pylint: disable=exec-used
            exec(file.read(), namespace)

        settings: InputBuildSettings | None = namespace.get('SETTINGS')
        rules: list[Rule] | None = namespace.get('RULES')
        custodian: Custodian | None = namespace.get('CUSTODIAN')
        context: Context | None = namespace.get('CONTEXT')
    else:
        label: str = f'-m {args.module.__name__}'
        settings: InputBuildSettings | None = getattr(args.module, 'SETTINGS', None)
        rules: list[Rule] | None = getattr(args.module, 'RULES', None)
        custodian: Custodian | None = getattr(args.module, 'CUSTODIAN', None)
        context: Context | None = getattr(args.module, 'CONTEXT', None)

    if args.audit_steps:
        audit_rules = context.rules if context else rules
        if not audit_rules:
            raise RuntimeError('Anchovy config files must have a RULES or CONTEXT attribute!')

        all_steps = set(Step.get_all_steps())
        available_steps = set(Step.get_available_steps())
        unavailable_steps = all_steps - available_steps
        used_steps = {r.step.__class__ for r in audit_rules if r.step}

        groups = {
            'Available steps': available_steps,
            'Unavailable steps': unavailable_steps,
            'Used steps': used_steps,
        }
        for group_label, step_group in groups.items():
            print(f'{group_label} ({len(step_group)})')
            for step in step_group:
                pprint_step(step)

    elif context or rules:
        try:
            if context:
                context.run()
            elif rules:
                run_from_rules(settings, rules, custodian, argv=remaining, prog=f'anchovy {label}')
        except StepUnavailableException as e:
            pprint_missing_deps(e.step)
            sys.exit(1)

    else:
        print_with_style(
            'Anchovy config files must have a RULES or CONTEXT attribute!',
            file='stderr',
            style='red'
        )
        sys.exit(1)



"""
Internal utilities for progress bars and pretty printing.
"""
import sys
import typing as t

try:
    import rich.progress as _rich_progress
except ImportError:
    _rich_progress = None
try:
    import rich.console
    _rich_consoles = {
        'stdout': rich.console.Console(file=sys.stdout),
        'stderr': rich.console.Console(file=sys.stderr),
    }
except ImportError:
    _rich_consoles = {}
try:
    import tqdm as _tqdm
except ImportError:
    _tqdm = None


T = t.TypeVar('T')


def track_progress(iterable: t.Iterable[T], desc: str) -> t.Iterable[T]:
    """
    Progress tracker which supports rich and tqdm progress bars and gracefully
    devolves to no progress tracking.
    """
    if _rich_progress:
        yield from _rich_progress.track(iterable, desc, console=_rich_consoles['stdout'])
    elif _tqdm:
        yield from _tqdm.tqdm(iterable, desc)
    else:
        print(desc)
        yield from iterable


def print_with_style(*args, sep=' ', end='\n', file: str = 'stdout', style=None):
    """
    Enhanced print() function which supports rich console styles and gracefully
    devolves to standard print().
    """
    if file in _rich_consoles:
        _rich_consoles[file].print(*args, sep=sep, end=end, style=style)
    else:
        print(*args, sep=sep, end=end, file=getattr(sys, file))



"""
Steps for working with Jinja templates, especially rendering Markdown into
Jinja templates.
"""
from __future__ import annotations

import shutil
import sys
import typing as t
from functools import reduce
from pathlib import Path

from .core import Step
from .dependencies import PipDependency, Dependency

if t.TYPE_CHECKING:
    from collections.abc import Sequence
    from jinja2 import Environment
    from markdown_it.renderer import RendererHTML
    from markdown_it.token import Token
    from markdown_it.utils import EnvType, OptionsDict

MDProcessor = t.Callable[[str], str]
MDContainerRenderer =  t.Callable[
    ['RendererHTML', 'Sequence[Token]', int, 'OptionsDict', 'EnvType'],
    str
]


class JinjaRenderStep(Step):
    """
    Abstract base class for Steps using Jinja rendering.
    """
    encoding = 'utf-8'

    @classmethod
    def get_dependencies(cls):
        return {
            PipDependency('jinja2'),
        }

    def __init__(self,
                 env: Environment | None = None,
                 extra_globals: dict[str, t.Any] | None = None):
        if env and extra_globals:
            env.globals.update(extra_globals)
        self._env = env
        self._extra_globals = extra_globals

    @property
    def env(self):
        """
        Returns the Jinja `Environment` for this Step, creating and caching it
        if necessary.
        """
        if self._env:
            return self._env

        from jinja2 import Environment, FileSystemLoader, select_autoescape
        self._env = Environment(
            loader=FileSystemLoader(self.context['input_dir']),
            autoescape=select_autoescape()
        )
        if self._extra_globals:
            self._env.globals.update(self._extra_globals)
        return self._env

    def render_template(self, template_name: str, meta: dict[str, t.Any], output_paths: list[Path]):
        """
        Render a Jinja template.

        :param template_name: The name of the template to render.
        :param meta: Any parameters to be passed to the template.
        :param output_paths: A list of Paths the rendered template will be
            written to.
        """
        if not output_paths:
            return

        template = self.env.get_template(template_name)
        for path in output_paths:
            path.parent.mkdir(parents=True, exist_ok=True)
        template.stream(**meta).dump(str(output_paths[0]), encoding=self.encoding)
        for path in output_paths[1:]:
            shutil.copy(output_paths[0], path)

        return template.filename


class JinjaMarkdownStep(JinjaRenderStep):
    """
    A Step for rendering Markdown using Jinja templates. Parses according to
    CommonMark and renders to HTML by default.
    """
    @classmethod
    def _build_markdownit(cls):
        import markdown_it
        processor = markdown_it.MarkdownIt()

        def convert(md_string: str) -> str:
            return processor.render(md_string)

        return convert

    @classmethod
    def _build_mistletoe(cls):
        import mistletoe

        def convert(md_string: str) -> str:
            return mistletoe.markdown(md_string)

        return convert

    @classmethod
    def _build_markdown(cls):
        import markdown
        processor = markdown.Markdown()

        def convert(md_string: str):
            return processor.convert(md_string)

        return convert

    @classmethod
    def _build_commonmark(cls):
        import commonmark
        parser = commonmark.Parser()
        renderer = commonmark.HtmlRenderer()

        def convert(md_string: str) -> str:
            return renderer.render(parser.parse(md_string))

        return convert

    @classmethod
    def get_options(cls):
        """
        Helper method returning a list of tuples of dependencies and markdown
        renderer factories for those dependencies.
        """
        return [
            (PipDependency('markdown-it-py', check_name='markdown_it'), cls._build_markdownit),
            (PipDependency('mistletoe'), cls._build_mistletoe),
            (PipDependency('markdown'), cls._build_markdown),
            (PipDependency('commonmark'), cls._build_commonmark),
        ]

    @classmethod
    def get_dependencies(cls):
        deps = [option[0] for option in cls.get_options()]
        dep_set = {reduce(lambda x, y: x | y, deps)} if deps else set[Dependency]()

        return super().get_dependencies() | dep_set

    def __init__(self,
                 default_template: str | None = None,
                 md_processor: MDProcessor | None = None,
                 jinja_env: Environment | None = None,
                 jinja_globals: dict[str, t.Any] | None = None):
        super().__init__(jinja_env, jinja_globals)
        self.default_template = default_template
        self._md_processor = md_processor

    @property
    def md_processor(self):
        """
        Returns the markdown processor for this Step, creating it if necessary.
        """
        if not self._md_processor:
            for dep, factory in self.get_options():
                if dep.satisfied:
                    self._md_processor = factory()
                    break
            else:
                raise RuntimeError('Markdown processor could not be initialized!')
        return self._md_processor


    def __call__(self, path: Path, output_paths: list[Path]):
        meta, content = self.extract_metadata(path.read_text(self.encoding))
        meta |= {'rendered_markdown': self.md_processor(content.strip()).strip()}

        template_path = self.render_template(
            meta.get('template', self.default_template),
            meta,
            output_paths
        )
        if template_path:
            return [path, Path(template_path)], output_paths

    def extract_metadata(self, text: str):
        """
        Read metadata from the front of a markdown-formatted text.
        """
        meta = {}
        lines = text.splitlines()

        i = 0
        for line in lines:
            if ':' not in line:
                break
            key, value = line.split(':', 1)
            if not key.isidentifier():
                break

            meta[key.strip()] = value.strip()
            i += 1

        return meta, '\n'.join(lines[i:])


class JinjaExtendedMarkdownStep(JinjaRenderStep):
    """
    A Step for extended Markdown rendering.

    Goes beyond the default functionality of markdown-it-py to offer toml
    frontmatter, pygments syntax highlighting for code blocks, containers,
    variable substitutions, wordcounts, anchors, and typography.
    """
    encoding = 'utf-8'

    @classmethod
    def get_dependencies(cls):
        deps = super().get_dependencies() | {
            PipDependency('markdown-it-py', check_name='markdown_it'),
            PipDependency('mdit_py_plugins'),
            PipDependency('Pygments', check_name='pygments'),
        }
        if sys.version_info < (3, 11):
            deps.add(PipDependency('tomli'))
        return deps

    def __init__(self,
                 default_template: str | None = None,
                 jinja_env: Environment | None = None,
                 jinja_globals: dict[str, t.Any] | None = None,
                 *,
                 container_types: list[tuple[str | None, list[str]]] | None = None,
                 container_renderers: dict[str, MDContainerRenderer] | None = None,
                 substitutions: dict[str, str] | None = None,
                 auto_anchors: bool = False,
                 auto_typography: bool = True,
                 code_highlighting: bool = True,
                 pygments_params: dict[str, t.Any] | None = None,
                 wordcount: bool = False):
        """
        :param default_template: The name of a Jinja template to use with
            markdown files that do not specify a template in their frontmatter.
        :param jinja_env: A custom Jinja2 `Environment`. A reasonable default
            will be provided if not specified.
        :param jinja_globals: Any parameters to be passed to the Jinja template.
            Additionally, all frontmatter keys will be included, rendered
            markdown will be added as `'rendered_markdown'`, and wordcount data
            will be added as a `'wordcount'` dict if enabled.
        :param container_types: A list of tuples pairing a HTML tag or None
            with a list of container names that should render to that HTML tag.
            If the HTML tag is none, the default raw `<div>` renderer from
            `mdit_py_plugins.container` will be used.
        :param container_renderers: A dictionary with container name as keys and
            container renderer functions as values, when additional processing
            is needed beyond the default options.
        :param substitutions: A dictionary of variable names and values to
            substitute into markdown before it is rendered. See
            `JinjaExtendedMarkdownStep.apply_substitutions()` for more details.
        :param auto_anchors: Whether to enable the `mdit_py_plugins.anchors`
            plugin.
        :param auto_typography: Whether to enable smartquotes and replacement
            functionalities in markdown-it-py.
        :param code_highlighting: Whether to enable code highlighting.
        :param pygments_params: Parameters to supply to
            `pygments.formatters.html.HtmlFormatter`.
        :param wordcount: Whether to enable the `mdit_py_plugins.wordcount`
            plugin.
        """
        super().__init__(jinja_env, jinja_globals)
        self.default_template = default_template
        self.container_types = container_types or []
        self.container_renderers = container_renderers or {}
        self.substitutions = substitutions or {}
        self.auto_anchors = auto_anchors
        self.auto_typography = auto_typography
        self.code_highlighting = code_highlighting
        self.pygments_params = pygments_params or {}
        self.wordcount = wordcount
        self._md_processor: t.Callable[[str], tuple[str, dict[str, t.Any]]] | None = None

    def __call__(self, path: Path, output_paths: list[Path]):
        rendered_md, meta = self.md_processor(
            self.apply_substitutions(
                path.read_text(self.encoding).strip()
            )
        )

        meta['rendered_markdown'] = rendered_md

        template_path = self.render_template(
            meta.get('template', self.default_template),
            meta,
            output_paths
        )
        if template_path:
            return [path, Path(template_path)], output_paths

    @property
    def md_processor(self):
        """
        Returns the markdown processor for this Step, creating it if necessary.
        """
        if not self._md_processor:
            self._md_processor = self._build_processor()
        return self._md_processor

    def apply_substitutions(self, text: str):
        """
        Apply variable substitutions to a markdown string.

        Looks for `${{ var_name }}`.
        """
        for sub, value in self.substitutions.items():
            text = text.replace('${{ ' + sub + ' }}', value)
        return text

    def highlight_code(self, code: str, lang: str, _lang_attrs: str):
        """
        Apply pygments syntax highlighting to the provided code, returning as
        HTML markup.
        """
        from pygments import highlight
        from pygments.formatters.html import HtmlFormatter
        from pygments.lexers import get_lexer_by_name, guess_lexer
        from pygments.util import ClassNotFound
        try:
            lexer = get_lexer_by_name(lang)
        except ClassNotFound:
            try:
                lexer = guess_lexer(code)
            except ClassNotFound:
                return ''

        return highlight(code, lexer, HtmlFormatter(**self.pygments_params))

    def _build_processor(self):
        import markdown_it
        # TODO Need for pyright suppression will be eliminated in the next
        # release of mdit_py_plugins:
        #  https://github.com/executablebooks/mdit-py-plugins/pull/91
        from mdit_py_plugins.anchors import anchors_plugin  # type: ignore[reportPrivateImportUsage]
        from mdit_py_plugins.attrs import attrs_block_plugin, attrs_plugin  # type: ignore[reportPrivateImportUsage]
        from mdit_py_plugins.container import container_plugin  # type: ignore[reportPrivateImportUsage]
        from mdit_py_plugins.front_matter import front_matter_plugin  # type: ignore[reportPrivateImportUsage]
        from mdit_py_plugins.wordcount import wordcount_plugin  # type: ignore[reportPrivateImportUsage]
        from .components import md_rendering

        processor = markdown_it.MarkdownIt(
            'commonmark',
            {
                'typographer': self.auto_typography,
                'highlight': self.highlight_code if self.code_highlighting else None,
            },
            renderer_cls=md_rendering.AnchovyRendererHTML
        )
        processor.enable(['strikethrough', 'table'])
        if self.auto_typography:
            processor.enable(['smartquotes', 'replacements'])
        if self.auto_anchors:
            anchors_plugin(processor)
        attrs_plugin(processor)
        attrs_block_plugin(processor)
        front_matter_plugin(processor)
        if self.wordcount:
            wordcount_plugin(processor)

        for tag, names in self.container_types:
            for name in names:
                renderer = (
                    self.container_renderers.get(name)
                    or (md_rendering.get_container_renderer(name, tag) if tag else None)
                )
                container_plugin(processor, name, render=renderer)

        def convert(md_string: str):
            env = {'anchovy_meta': dict[str, t.Any]()}
            rendered_md: str = processor.render(md_string, env=env)
            meta = env['anchovy_meta']
            if self.wordcount:
                meta['wordcount'] = env['wordcount']
            return rendered_md, meta

        return convert



"""
Anchovy is a minimal, unopinionated file-processing framework equipped with a
complete static website generation toolkit.
"""
from .core import Context, InputBuildSettings, Matcher, PathCalc, Rule, Step
from .css import AnchovyCSSStep
from .custody import Custodian, CustodyEntry
from .dependencies import Dependency, PipDependency, WebExecDependency
from .images import CWebPStep, ImageMagickStep, IMThumbnailStep, PillowStep, OptipngStep
from .include import RequestsFetchStep, UnpackArchiveStep, URLLibFetchStep
from .jinja import JinjaExtendedMarkdownStep, JinjaMarkdownStep, JinjaRenderStep
from .minify import CSSMinifierStep, HTMLMinifierStep, ResourcePackerStep
from .paths import DirPathCalc, OutputDirPathCalc, REMatcher, WorkingDirPathCalc
from .simple import DirectCopyStep



"""
Steps for image optimization, conversion, and transformation.
"""
from __future__ import annotations

import shutil
import typing as t
from pathlib import Path

from .core import Step
from .dependencies import PipDependency, WebExecDependency
from .simple import BaseCommandStep

if t.TYPE_CHECKING:
    from _typeshed import StrOrBytesPath


class CWebPStep(BaseCommandStep):
    """
    A WebP image conversion/optimization Step using cwebp.
    """
    def __init__(self,
                 quality: int = 75,
                 lossless: bool = False,
                 options: t.Iterable[str] = ()):
        """
        @quality controls the size of the output image, traded off with quality
        for lossy images and processing time for lossless images. @lossless can
        be used to avoid image degradation, at the cost of file size. See
        https://developers.google.com/speed/webp/docs/cwebp for further @options
        and explanations.
        """
        self.options = ['-q', str(quality)]
        if lossless:
            self.options.append('-lossless')
        self.options.extend(options)

    @classmethod
    def get_dependencies(cls):
        return {
            WebExecDependency('cwebp', 'https://developers.google.com/speed/webp/download'),
        }

    def get_command(self, input_path: Path, output_path: Path) -> list[StrOrBytesPath]:
        return ['cwebp', input_path, *self.options, '-o', output_path]


class ImageMagickStep(BaseCommandStep):
    """
    A completely customizable image transformation Step using ImageMagick.
    """
    def __init__(self, options: t.Iterable[str] = ()):
        """
        If simple image conversion is desired, @options can be left empty.
        Otherwise, see https://imagemagick.org/script/command-line-options.php,
        or perhaps use `IMThumbnailStep()` instead.
        """
        self.options = options

    @classmethod
    def get_dependencies(cls):
        return {
            WebExecDependency(
                'imagemagick',
                'https://imagemagick.org/script/download.php',
                'magick'
            ),
        }

    def get_command(self, input_path: Path, output_path: Path) -> list[StrOrBytesPath]:
        return ['magick', input_path, *self.options, output_path]


class IMThumbnailStep(ImageMagickStep):
    """
    A more specialized ImageMagick Step intended for thumbnailing.
    """
    def __init__(self,
                 dimensions: str = '300x300',
                 fill_color: str | None = None,
                 extra_options: t.Iterable[str] = ()):
        """
        @dimensions can be any "geometry" supported by imagemagick, and sets
        the target thumbnail size. Aspect ratio will be preserved. If output
        images need uniform dimensions, @fill_color can be supplied to fill the
        extra space; use something like '#0000' for transparent padding.
        Additional custom options may be passed using @extra_options.
        """
        options = ['-resize', dimensions]
        if fill_color:
            # https://legacy.imagemagick.org/discourse-server/viewtopic.php?t=26971
            options.extend([
                '-background', fill_color,
                '-gravity', 'center',
                '-extent', dimensions
            ])
        options.extend(extra_options)
        super().__init__(options)


class PillowStep(Step):
    """
    A simple Pillow step which can convert and/or thumbnail images.
    """
    def __init__(self, thumbnail: tuple[int, int] | None = None):
        self.thumbnail = thumbnail

    @classmethod
    def get_dependencies(cls):
        return {
            PipDependency('Pillow', check_name='PIL'),
        }

    def __call__(self, path: Path, output_paths: list[Path]):
        from PIL import Image

        groups = {}
        for target_path in output_paths:
            groups.setdefault(target_path.suffix, []).append(target_path)

        with Image.open(path) as img:
            if self.thumbnail:
                img.thumbnail(self.thumbnail)

            for first, *remainder in groups.values():
                first.parent.mkdir(parents=True, exist_ok=True)
                img.save(first)
                for dup in remainder:
                    dup.parent.mkdir(parents=True, exist_ok=True)
                    shutil.copy(first, dup)


class OptipngStep(BaseCommandStep):
    """
    A PNG optimization step using optipng.
    """
    def __init__(self,
                 optimization_level: int | None = None,
                 extra_options: t.Iterable[str] = ()):
        """
        The default @optimization_level may vary based on your build of optipng
        but is probably 2. Extra flags may be supplied using @extra_options.
        """
        self.options = ['-o', str(optimization_level)] if optimization_level is not None else []
        self.options.extend(extra_options)

    @classmethod
    def get_dependencies(cls):
        return {
            WebExecDependency('optipng', 'http://optipng.sourceforge.net'),
        }

    def get_command(self, input_path: Path, output_path: Path) -> list[StrOrBytesPath]:
        return ['optipng', *self.options, input_path, '-out', output_path]



"""
Core classes and types for the Anchovy processing pipeline.
"""
from __future__ import annotations

import abc
import shutil
import typing as t
from pathlib import Path

from .custody import CustodyEntry, Custodian
from .dependencies import Dependency
from .pretty_utils import track_progress


T = t.TypeVar('T')
T2 = t.TypeVar('T2')
ContextDir = t.Literal['input_dir', 'output_dir', 'working_dir']
BuildSettingsKey = t.Literal[ContextDir, 'custody_cache', 'purge_dirs']


class InputBuildSettings(t.TypedDict, total=False):
    """
    TypedDict for defining build settings in an Anchovy config file.
    """
    input_dir: Path
    output_dir: Path
    working_dir: Path | None
    custody_cache: Path | None
    purge_dirs: bool | None

class BuildSettings(t.TypedDict):
    """
    TypedDict for processed build settings ready for passing to Context.
    """
    input_dir: Path
    output_dir: Path
    working_dir: Path
    custody_cache: Path | None
    purge_dirs: bool | None


def _rm_children(path: Path):
    if not path.exists():
        return
    for child in path.iterdir():
        if child.is_dir():
            shutil.rmtree(child)
        else:
            child.unlink()


def _rm_orphans(path: Path, exclude: set[Path]):
    if not path.exists():
        return False
    removed_all = True
    for child in path.iterdir():
        if child in exclude:
            removed_all = False
            continue
        if child.is_dir():
            if _rm_orphans(child, exclude):
                child.rmdir()
            else:
                removed_all = False
        else:
            child.unlink()
    return removed_all


class Context:
    """
    A context and configuration class for building Anchovy projects.
    """
    def __init__(self,
                 settings: BuildSettings,
                 rules: list[Rule],
                 custodian: Custodian | None = None):
        self.settings = settings
        # Rules may need access to the Custodian
        self.custodian = custodian or Custodian()
        self.rules: list[Rule] = []
        for rule in rules:
            self.rules.append(rule)
            self.bind(rule.step)

    @t.overload
    def __getitem__(self, key: ContextDir) -> Path: ...
    @t.overload
    def __getitem__(self, key: t.Literal['custody_cache']) -> Path | None: ...
    @t.overload
    def __getitem__(self, key: t.Literal['purge_dirs']) -> bool | None: ...
    def __getitem__(self, key):
        return self.settings[key]

    def bind(self, step: Step | None):
        """
        Bind a Step to this Context, checking to ensure its availability.
        """
        if step:
            if not step.is_available():
                raise StepUnavailableException(step)
            step.bind(self)

    def find_inputs(self, path: Path):
        """
        Overridable function to get paths to process based on a given @path.
        Default behavior is to recursively search for files but exclude the
        directories themselves.
        """
        for candidate in path.iterdir():
            if candidate.is_dir():
                yield from self.find_inputs(candidate)
            else:
                yield candidate

    def process(self, input_paths: list[Path] | None = None):
        """
        Process a set of files using the Context's defined rules. If
        @input_paths is empty or None, `self.find_inputs()` will be used to get
        a tree of files to process. If intermediate files are produced,
        `self.process()` will be called recursively with them.
        """
        input_paths = input_paths or list(self.find_inputs(self.settings['input_dir']))
        # We want to handle tasks in the order they're defined!
        tasks: dict[Step, list[tuple[Path, list[Path]]]]
        tasks = {r.step: [] for r in self.rules if r.step}

        for path in track_progress(input_paths, 'Planning...'):
            for rule in self.rules:
                if match := rule.matcher(self, path):
                    # None can be used to halt further rule processing.
                    if not rule.step:
                        break
                    output_paths: list[Path] = []
                    for pathcalc in rule.path_calcs:
                        # None can be used to halt further rule processing in
                        # paths as well. This allows a single rule to both do
                        # processing and also halt further processing.
                        if not pathcalc:
                            break
                        output_paths.append(pathcalc(self, path, match))
                    else:
                        tasks[rule.step].append((path, output_paths))
                        # We didn't break above, avoid the break below!
                        continue
                    tasks[rule.step].append((path, output_paths))
                    # We need two breaks because we're trying to get out of the
                    # surrounding for loop.
                    break

        flattened: list[tuple[Step, Path, list[Path]]] = []
        for step, paths in tasks.items():
            flattened.extend((step, p, ops) for p, ops in paths)

        further_processing: list[Path] = []
        for step, path, output_paths in track_progress(flattened, 'Processing...'):
            stale, msg = self.custodian.refresh_needed(path, output_paths)
            if stale:
                explicit_chain = step(path, output_paths)
                if explicit_chain:
                    sources, output_paths = explicit_chain
                else:
                    sources = [path]
                self.custodian.add_step(sources, output_paths, msg)
            else:
                output_paths = self.custodian.skip_step(path, output_paths)

            further_processing.extend(
                p for p in output_paths
                if p.is_relative_to(self['working_dir'])
            )

        if further_processing:
            self.process(further_processing)

    def run(self, input_paths: list[Path] | None = None):
        """
        Execute pre-run hooks (currently only the baked-in directory purge),
        then call `self.process()` with @input_paths.
        TODO: Support custom pre/post hooks.
        """
        if self['purge_dirs']:
            _rm_children(self['output_dir'])
            _rm_children(self['working_dir'])
        self.custodian.bind(self)
        if cache_file := self['custody_cache']:
            self.custodian.load_file(cache_file)
        self.process(input_paths)
        if cache_file:
            self.custodian.dump_file(cache_file)
            if self['purge_dirs'] is None:
                touched = set(self.custodian.get_all_paths())
                _rm_orphans(self['output_dir'], touched)
                _rm_orphans(self['working_dir'], touched)


class Matcher(t.Generic[T], abc.ABC):
    """
    Abstract base class for Path Matchers. Provides pre-baked ability to
    combine Matchers with | and &.
    """
    @abc.abstractmethod
    def __call__(self, context: Context, path: Path) -> T:
        ...

    def __or__(self, other: Matcher[T2]):
        return _OrMatcher(self, other)

    def __and__(self, other: Matcher[T2]):
        return _AndMatcher(self, other)


class _OrMatcher(Matcher[T | T2]):
    def __init__(self, left: Matcher[T], right: Matcher[T2]):
        self.left = left
        self.right = right

    def __call__(self, context: Context, path: Path):
        return self.left(context, path) or self.right(context, path)


class _AndMatcher(Matcher[T | T2]):
    def __init__(self, left: Matcher[T], right: Matcher[T2]):
        self.left = left
        self.right = right

    def __call__(self, context: Context, path: Path):
        return self.left(context, path) and self.right(context, path)


class PathCalc(t.Generic[T], abc.ABC):
    """
    Abstract base class for path calculators which use `Matcher` match data to
    determine output paths from input paths.
    """
    @abc.abstractmethod
    def __call__(self, context: Context, path: Path, match: T) -> Path:
        ...


class Rule(t.Generic[T]):
    """
    A single rule for Anchovy file processing, with a matcher, output path
    calculators, and an optional Step to run.
    """
    def __init__(self,
                 matcher: Matcher[T],
                 path_calc: t.Sequence[PathCalc[T] | Path | None] | PathCalc[T] | Path | None,
                 step: Step | None = None):
        self.matcher = matcher
        self.step = step
        if not isinstance(path_calc, t.Sequence):
            path_calc = [path_calc]
        self.path_calcs = [self._path_to_pathcalc(p) if isinstance(p, Path) else p for p in path_calc]

    def _path_to_pathcalc(self, path: Path):
        # pylint: disable=unused-argument
        def wrapper(context: Context, input_path: Path, match: t.Any):
            return path
        return wrapper


class Step(abc.ABC):
    """
    Abstract base class for Steps, individual processing stages used to build a
    full Anchovy ruleset.
    """
    context: Context
    _step_registry: list[t.Type[Step]] = []

    def __init_subclass__(cls, **kw):
        super().__init_subclass__(**kw)
        cls._step_registry.append(cls)

    @classmethod
    def get_all_steps(cls):
        """
        Return a list of all currently known Steps.
        """
        return list(cls._step_registry)

    @classmethod
    def get_available_steps(cls):
        """
        Return a list of all currently known Steps whose requirements are met.
        """
        return [s for s in cls._step_registry if s.is_available()]

    @classmethod
    def is_available(cls) -> bool:
        """
        Return whether this Step's requirements are installed, making it
        available for use.
        """
        return all(d.satisfied for d in cls.get_dependencies())

    @classmethod
    def get_dependencies(cls) -> set[Dependency]:
        """
        Return the requirements for this Step.
        """
        return set()

    def bind(self, context: Context):
        """
        Bind this Step to a Context.
        """
        self.context = context

    @abc.abstractmethod
    def __call__(
        self,
        path: Path,
        output_paths: list[Path]
    ) -> None | tuple[list[Path | CustodyEntry], list[Path]]:
        ...


class StepUnavailableException(Exception):
    """
    Exception raised with a step to be used is unavailable due to missing
    dependencies.
    """
    def __init__(self, step: Step, *args: t.Any):
        self.step = step
        super().__init__(*args)



"""
Practical implementations of Matchers and PathCalcs.
"""
import re
import typing as t
from pathlib import Path

from .core import Context, ContextDir, Matcher, PathCalc


T = t.TypeVar('T')


def _trim_ext_prefix(path: Path, match: re.Match[str]):
    _groups = match.groupdict()
    if 'stem' in _groups:
        return path.with_stem(_groups['stem'])
    if 'ext' in _groups and _groups['ext']:
        return path.with_name(path.name[:-len(_groups['ext'])])
    return path


def _to_dir_inner(dest: Path,
                  ext: str | None,
                  context: Context,
                  path: Path,
                  match: t.Any,
                  transform: t.Callable[[Path], Path] | None = None):
    path = _trim_ext_prefix(path, match) if ext and isinstance(match, re.Match) else path

    rel = path.relative_to(
        context['input_dir']
        if path.is_relative_to(context['input_dir'])
        else context['working_dir']
    )
    if transform:
        rel = transform(rel)
    new_path = dest / rel

    if ext is not None:
        new_path = new_path.with_suffix(ext)

    return new_path


class DirPathCalc(PathCalc[T]):
    """
    PathCalc which makes its input paths children of a specified directory.
    If @ext is specified, it will replace the extension of input paths. If the
    matcher produced an re.Match, it will be checked for explicitly defined
    extension information for the input paths, allowing for meaningful work
    with extensions that `pathlib.Path` does not reflect, like `.tar.gz`.
    """
    def __init__(self, dest: Path, ext: str | None = None, transform: t.Callable[[Path], Path] | None = None):
        self.dest = dest
        self.ext = ext
        self.transform = transform

    def __call__(self, context: Context, path: Path, match: T) -> Path:
        return _to_dir_inner(self.dest, self.ext, context, path, match, self.transform)


class OutputDirPathCalc(PathCalc[T]):
    """
    PathCalc which makes its input paths children of the Context's output
    directory. If @ext is specified, it will replace the extension of input
    paths. If the matcher produced an re.Match, it will be checked for
    explicitly defined extension information for the input paths, allowing for
    meaningful work with extensions that `pathlib.Path` does not reflect, like
    `.tar.gz`.
    """
    def __init__(self, ext: str | None = None, transform: t.Callable[[Path], Path] | None = None):
        self.ext = ext
        self.transform = transform

    def __call__(self, context: Context, path: Path, match: T) -> Path:
        return _to_dir_inner(context['output_dir'], self.ext, context, path, match, self.transform)


class WorkingDirPathCalc(PathCalc[T]):
    """
    PathCalc which makes its input paths children of the Context's working
    directory. If @ext is specified, it will replace the extension of input
    paths. If the matcher produced an re.Match, it will be checked for
    explicitly defined extension information for the input paths, allowing for
    meaningful work with extensions that `pathlib.Path` does not reflect, like
    `.tar.gz`.
    """
    def __init__(self, ext: str | None = None, transform: t.Callable[[Path], Path] | None = None):
        self.ext = ext
        self.transform = transform

    def __call__(self, context: Context, path: Path, match: T) -> Path:
        return _to_dir_inner(context['working_dir'], self.ext, context, path, match, self.transform)


class REMatcher(Matcher[re.Match | None]):
    """
    Path Matcher using regular expressions. @re_flags will be passed to
    `re.compile()`. @parent_dir, if specified, should be a key to a configured
    directory, not a Path, and will be used to handle matching the beginning of
    Paths; this can be used to avoid pitfalls with unexpected characters in
    input or working directories.
    """
    def __init__(self, re_string: str, re_flags: int = 0, parent_dir: ContextDir | None = None):
        self.regex = re.compile(re_string, re_flags)
        self.parent_dir: ContextDir | None = parent_dir

    def __call__(self, context: Context, path: Path):
        if self.parent_dir:
            # Handle this part of matching outside the regex.
            if not path.is_relative_to(context[self.parent_dir]):
                return None
            path = path.relative_to(context[self.parent_dir])
        return self.regex.match(path.as_posix())



"""
Dependency management for Anchovy Steps.
"""
from __future__ import annotations

import abc
import importlib
import shutil


class Dependency(abc.ABC):
    """
    A base class for trackable, evaluable, composable dependencies.
    """

    @property
    @abc.abstractmethod
    def satisfied(self) -> bool:
        """
        A bool indicating whether this dependency is met.
        """

    @property
    def needed(self) -> bool:
        """
        A bool indicating whether this dependency is needed on the current platform.
        """
        return True

    @property
    @abc.abstractmethod
    def install_hint(self) -> str:
        """
        A string giving help on how to install this dependency.
        """

    def __repr__(self):
        return f'{self.__class__.__name__}({self}, needed={self.needed}, satisfied={self.satisfied})'

    def __or__(self, other: Dependency):
        return _OrDependency(self, other)

    def __and__(self, other: Dependency):
        return _AndDependency(self, other)


class _OrDependency(Dependency):
    def __init__(self, left: Dependency, right: Dependency):
        self.left = left
        self.right = right

    def __repr__(self):
        return f'({self.left!r} | {self.right!r})'

    def __str__(self):
        return f'({self.left} | {self.right})'

    @property
    def satisfied(self):
        """
        A bool indicating whether either of these dependencies are met.
        """
        return self.left.satisfied or self.right.satisfied

    @property
    def needed(self):
        """
        A bool indicating whether either of these dependencies is needed.
        """
        return self.left.needed or self.right.needed

    @property
    def install_hint(self):
        """
        A string giving help on how to meet one of these dependencies, if
        either is needed.
        """
        return (
            (self.left.needed and self.left.install_hint)
            or (self.right.needed and self.right.install_hint)
            or ''
        )


class _AndDependency(Dependency):
    def __init__(self, left: Dependency, right: Dependency):
        self.left = left
        self.right = right

    def __repr__(self):
        return f'({self.left!r} & {self.right!r})'

    def __str__(self):
        return f'({self.left} & {self.right})'

    @property
    def satisfied(self):
        """
        A bool indicating whether both of these dependencies are met.
        """
        return self.left.satisfied and self.right.satisfied

    @property
    def needed(self):
        """
        A bool indicating whether either of these dependencies is needed.
        """
        # These should mostly match, but we'll allow for cases where one
        # dependency is paired on more specific platforms.
        return self.left.needed or self.right.needed

    @property
    def install_hint(self):
        """
        A string giving help on how to install these dependencies, if they are
        needed.
        """
        return '; '.join(d.install_hint for d in [self.left, self.right] if d.needed)


class PipDependency(Dependency):
    """
    A Dependency on a pip-installable package.
    """
    def __init__(self,
                 name: str,
                 source: str | None = None,
                 check_name: str | None = None):
        self.name = name
        self.source = source or name
        self.check_name = check_name or name

    def __str__(self):
        return self.name

    @property
    def satisfied(self):
        """
        A bool indicating whether this dependency is met.
        """
        try:
            importlib.import_module(self.check_name)
        except ImportError:
            return False
        return True

    @property
    def install_hint(self):
        """
        A string giving help on how to install this dependency.
        """
        return f'pip install {self.source}'


class WebExecDependency(Dependency):
    """
    A Dependency on a general internet-sourced executable.
    """
    def __init__(self,
                 name: str,
                 source: str | None = None,
                 check_name: str | None = None):
        self.name = name
        self.source = source or name
        self.check_name = check_name or name

    def __str__(self):
        return self.name

    @property
    def satisfied(self):
        """
        A bool indicating whether this dependency is met.
        """
        return bool(shutil.which(self.check_name))

    @property
    def install_hint(self):
        """
        A string giving help on how to install this dependency.
        """
        return self.source



"""
Steps for pre-processing the Anchovy CSS format.
"""
from __future__ import annotations

from pathlib import Path

from .core import Step
from .dependencies import PipDependency


class AnchovyCSSStep(Step):
    """
    Simple Step to preprocess an Anchovy CSS file into compliant CSS.
    """
    encoding = 'utf-8'

    @classmethod
    def get_dependencies(cls):
        return {
            PipDependency('anchovy_css'),
        }

    def __call__(self, path: Path, output_paths: list[Path]):
        if not output_paths:
            return
        from anchovy_css import process
        processed = process(path.read_text(self.encoding))
        for target_path in output_paths:
            target_path.parent.mkdir(parents=True, exist_ok=True)
            target_path.write_text(processed, self.encoding)



"""
Steps to add files from network resources and archives into an Anchovy build.
"""
from __future__ import annotations

import shutil
import sys
from pathlib import Path

from anchovy.core import Context

from .core import Step
from .custody import CustodyEntry
from .dependencies import PipDependency


class RequestsFetchStep(Step):
    """
    A step using requests to fetch a resource from a URL in a config file into
    the build.
    """
    chunk_size = 8192
    timeout = 300

    @classmethod
    def get_dependencies(cls):
        deps = {
            PipDependency('requests'),
        }
        if sys.version_info < (3, 11):
            deps.add(PipDependency('tomli'))
        return deps

    def bind(self, context: Context):
        super().bind(context)
        @context.custodian.register_checker('requests', override=False)
        def requests_resource_stale(entry: CustodyEntry):
            import requests
            response = requests.head(
                entry.key,
                allow_redirects=True,
                headers={'If-None-Match': entry['etag']},
                timeout=self.timeout
            )
            return response.status_code == 304

    def __call__(self, path: Path, output_paths: list[Path]):
        if not output_paths:
            return
        for o_path in output_paths:
            o_path.parent.mkdir(parents=True, exist_ok=True)

        import requests
        if sys.version_info < (3, 11):
            import tomli as tomllib
        else:
            import tomllib

        with path.open('rb') as file:
            config = tomllib.load(file)
        url: str = config.pop('url')
        config.setdefault('stream', True)
        config.setdefault('timeout', self.timeout)

        response = requests.get(url, **config)
        if not response:
            response.raise_for_status()
        with output_paths[0].open('wb') as file:
            for chunk in response.iter_content(self.chunk_size):
                file.write(chunk)
            for o_path in output_paths[1:]:
                shutil.copy(output_paths[0], o_path)

        centry = CustodyEntry('requests', url, {'etag': response.headers['ETag']})
        return [path, centry], output_paths


class URLLibFetchStep(Step):
    """
    A step using urllib to fetch a resource from a URL in a config file into
    the build.
    """
    @classmethod
    def get_dependencies(cls):
        return {PipDependency('tomli')} if sys.version_info < (3, 11) else {}

    def bind(self, context: Context):
        super().bind(context)
        @context.custodian.register_checker('urllib', override=False)
        def urllib_resource_stale(entry: CustodyEntry):
            import urllib.request
            request = urllib.request.Request(
                entry.key,
                headers={'If-None-Match': entry['etag']},
                method='HEAD'
            )
            try:
                with urllib.request.urlopen(request):
                    return False
            except urllib.request.HTTPError as e:
                if e.code == 304:
                    return True
                raise

    def __call__(self, path: Path, output_paths: list[Path]):
        if not output_paths:
            return
        for o_path in output_paths:
            o_path.mkdir(parents=True, exist_ok=True)

        import urllib.request
        if sys.version_info < (3, 11):
            import tomli as tomllib
        else:
            import tomllib

        with path.open('rb') as file:
            config = tomllib.load(file)
        url: str = config.pop('url')

        _path, msg = urllib.request.urlretrieve(url, output_paths[0], **config)

        for o_path in output_paths[1:]:
            shutil.copy(output_paths[0], o_path)

        centry = CustodyEntry('urllib', url, {'etag': msg['ETag']})
        return [path, centry], output_paths


class UnpackArchiveStep(Step):
    """
    A step for extracting files from an archive.
    """
    def __init__(self, archive_format: str | None = None):
        self.archive_format = archive_format
    def __call__(self, path: Path, output_paths: list[Path]):
        if not output_paths:
            return
        for o_path in output_paths:
            o_path.mkdir(parents=True, exist_ok=True)

        first = output_paths[0]
        shutil.unpack_archive(path, first, format=self.archive_format)
        all_outputs = [first]
        all_outputs.extend(self.context.find_inputs(first))
        for o_path in output_paths[1:]:
            shutil.copytree(first, o_path, dirs_exist_ok=True)
            all_outputs.append(o_path)
            all_outputs.extend(self.context.find_inputs(o_path))
        return [path], all_outputs
